//	    -------------------------------------
//	    OTHELLO II v1.45  par Franois LEIBER
//	        Francois.leiber@libertysurf.fr
//	             Compil avec Ti-gcc
//	   Librairies documentes par Zeljko Juric
//		       		  10/06/2000
//	    -------------------------------------

// For those who don't speak french and who'd like to understand my comments, just e-mail me, I'd be pleased to answer you
// if ever I have some free time :)

// Attention, ce programme n'est plus compatible avec le mode nostub : toutes les variables globales ne sont pas initialises
// et le programme ne s'occupe pas de restaurer l'cran au moment de quitter.


#include <doors.h>
#include <all.h>
#include "othello2.h"

int _ti89;

char _comment[] = fr_eng("Othello II v1.4b par Francois Leiber", "Othello II v1.4b by Francois Leiber");

BYTE ordre[60], tab_sav[61] = {}; 
int coeff[71], tourne_pion_y, nbre_pions, perdant = 0, taille;
BYTE tips = 1, faire_tourner = 1;
char jmp_quit[56], jmp_return[56], jmp_debut[56];

#include "graphiques.h"
#include "reflexion.h"

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Teste si un joueur peut jouer sur une case
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int test_case(TABLEAU tableau, int caze, int joueur)
{
	int horiz, vert, pas, i;
	if (tableau[caze] != 2) return 0;
	for (horiz=(caze%9>=2)?(-1):(0); horiz<=(caze%9<=5)?(1):(0); horiz++)
		for (vert=(caze>=18)?(-1):(0); vert<=(caze<=52)?(1):(0); vert++)
		{
			pas = horiz + 9*vert;
			if (tableau[caze+pas] != !joueur) continue;
			for (i=caze+2*pas; ((pas>0)?(i<=70):(i>=0))&&(tableau[i] < 2); i+=pas) if (tableau[i] == joueur) return 1;
		}
	return 0;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Fait voluer les coefficients et le plateau lorsque la case caze est joue
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void evol_tableau(TABLEAU tableau, int caze)
{
	static BYTE evol_coeff1[] = {0,0,0 ,7, 7, 7,63,63,63,70,70,70,2 ,3 ,4 ,5 ,18,27,36,45,25,34,43,52,65,66,67,68,2,3,4,5,18,27,25,34,36,45,43,52,65,66,67,68};
	static BYTE evol_coeff2[] = {1,9,10,6,15,16,54,55,64,60,61,69,11,12,13,14,19,28,37,46,24,33,42,51,56,57,58,59,1,1,6,6,9 ,9 ,16,16,54,54,61,61,64,64,69,69};
	int i, j;
	
	for (i=(caze%9>=1)?(-1):(0); i<=(caze%9<=6)?(1):(0); i++)		// Les cases vides autour du pion plac sont initialises  2
		for (j=(caze>=9)?(-1):(0); j<=(caze<=61)?(1):(0); j++) if (tableau[caze+9*j+i] == 3) tableau[caze+9*j+i] = 2;

	for (i=0; i<=43; i++) if (caze == evol_coeff1[i]) coeff[evol_coeff2[i]] = 30;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Charge le n-ime coup de la partie stocke dans tab_sav
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void charger(int coup, TABLEAU tableau, int *joueur, int *nbre_blancs)
{
	static int coeff_nouveau[71] = {	2000, -99,41,29,31,39, -98,2000,0,
										-98, -999, 5, 8, 8, 5,-999, -99,0,
										 40,    5,64,32,31,66,   5,  40,0,
										 30,    8,31,14,14,31,   8,  30,0,
										 30,    8,32,14,14,32,   8,  30,0,
										 40,    5,66,31,31,64,   5,  40,0,
										-99, -999, 5, 8, 8, 5,-999, -98,0,
										2000, -98,39,31,29,41, -99,2000		};
	static TABLEAU tab_nouveau = {	3,3,3,3,3,3,3,3,3,			// 0 = pion blanc
									3,3,3,3,3,3,3,3,3,			// 1 = pion noir
									3,3,2,2,2,2,3,3,3,			// 2 = case vide  ct d'un pion
									3,3,2,0,1,2,3,3,3,			// 3 = case vide loin de tout pion
									3,3,2,1,0,2,3,3,3,
									3,3,2,2,2,2,3,3,3,
									3,3,3,3,3,3,3,3,3,
									3,3,3,3,3,3,3,3		};
	int i, j, horiz, vert, pas, caze;

	*joueur = 1; *nbre_blancs = 2;
	memcpy(tableau, tab_nouveau, sizeof(TABLEAU));
	memcpy(coeff, coeff_nouveau, sizeof(coeff_nouveau));
	
	for (nbre_pions=4; nbre_pions<coup; nbre_pions++)			// Rejoue tous les coups depuis le dbut du jeu
	{
		caze = tab_sav[nbre_pions-4];
		if (caze == 255) break;
		tableau[caze] = *joueur;
		if (!*joueur) (*nbre_blancs)++;
		for (horiz=(caze%9>=2)?(-1):(0); horiz<=(caze%9<=5)?(1):(0); horiz++)		// Retourne les pions
			for (vert=(caze>=18)?(-1):(0); vert<=(caze<=52)?(1):(0); vert++)
			{
				pas = 9*vert+horiz;
				if (!(tableau[caze+pas] == !*joueur)) continue;
				for (i=caze+2*pas; (pas>0)?(i<=70):(i>=0); i+=pas)
				{
					if (tableau[i] >= 2) break;
					if (tableau[i] == *joueur)
					{
						for (j=caze+pas; (pas>0)?(j<=i-pas):(j>=i-pas); j+=pas)
						{
							tableau[j] = *joueur;
							*nbre_blancs += (*joueur)?(-1):(1);
						}
						break;
					}
				}
			}
		evol_tableau(tableau, caze);
		*joueur = !*joueur;
	}
	
	if (perdant) for(i=0; i<=70; i++) coeff[i] *= -1;
	*joueur = !*joueur;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "menus.h"

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Procdure appele au lancement, contient la majorit du jeu
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void jeu(void)
{
	static BYTE ordre_nouveau[60] = {0,7,63,70,2,5,18,25,45,52,65,68,20,23,47,50,3,4,27,34,36,43,66,67,21,22,29,32,38,41,48,49,12,13,28,33,37,42,57,58,11,14,19,24,46,51,56,59,1,6,9,16,54,61,64,69,10,15,55,60};
	static BYTE joueur_type[2] = {4, 0};
	int tourner[24];
	int i, j, touche, max, score, barre1, barre2, horiz, vert, pas;
	int nbre_blancs, caze = 30, caz, joueur, nbre_ordi, nbre;
	TABLEAU tab, tableau;
	LCD_BUFFER LCD_tmp;
 
	menu(tableau, joueur_type, &nbre_ordi, &joueur, &nbre_blancs, 0);		// Menu de prsentation

	setjmp(jmp_debut);											// Sert  pouvoir recommencer un nouveau jeu


	for (; nbre_pions<=63; nbre_pions++)			// Boucle principale
	{
		taille = 62-nbre_pions;
		Boucle:

		joueur = !joueur;

		setjmp(jmp_return);			// Permet de ressortir du menu sans recommencer le jeu

		printf_xy(97, 51, fr_eng("%d BLANCS ", "%d WHITES "), nbre_blancs);		// Affichage du nombre de pions de chaque couleur
		printf_xy(97, 37, fr_eng("%d NOIRS ", "%d BLACKS "), nbre_pions - nbre_blancs);
		DrawChar(90, 36, '', 2); DrawChar(90, 50, '', 2);
		LCD_save(LCD_tmp);

		for (i=0, touche=0; i<=59; i++) if (tableau[ordre_nouveau[i]] > 1) ordre[touche++] = ordre_nouveau[i];	// Evolution du tableau 'ordre' (cases  regarder)

		for (i=0, nbre=0; i<=taille+1; i++)								// Affiche un point sur les cases o l'on peut jouer
			if (test_case(tableau, caz = ordre[i], joueur))				// et stocke le numro de ces 'nbre' cases dans tourner
			{
				if (!joueur_type[joueur]&&tips) DrawPix(10*(caz%9)+7, 10*caz/9-10*caz%9+17, 1);
				tourner[nbre++] = caz;
			}
			if (!nbre)						 // Saute le tour du joueur s'il n'a pas de coup valide et termine le jeu si personne ne peut jouer
			{
				for (i=0; i<=taille+1; i++) if (test_case(tableau, ordre[i], !joueur)) nbre = 1; if (!nbre) break;
				nbre_pions--;
				if (nbre_ordi < 2)
				{
					ScrRectFill(&(SCR_RECT){19, 39, 139, 59}, &(SCR_RECT){0, 0, 159, 99}, 0);
					ScrRectFill(&(SCR_RECT){18, 40, 140, 58}, &(SCR_RECT){0, 0, 159, 99}, 0);
					ScrRectFill(&(SCR_RECT){20, 40, 138, 58}, &(SCR_RECT){0, 0, 159, 99}, 1);
					ScrRectFill(&(SCR_RECT){19, 41, 139, 57}, &(SCR_RECT){0, 0, 159, 99}, 1);
					DrawClipRect(&(WIN_RECT){21, 42, 137, 56}, &(SCR_RECT){0, 0, 159, 99}, 0);
					DrawStr(fr_eng(23, 38), 46, (nbre_ordi)?((joueur_type[joueur])?(fr_eng("La 89 doit passer !", "The 89 passes!")):(fr_eng("Vous devez passer !", "You must pass!"))):((joueur)?(fr_eng("Les noirs passent !", " Blacks pass!")):(fr_eng("Les blancs passent!", " Whites pass!"))), 2);
					ngetchx();
				}
				LCD_restore(LCD_tmp);
				continue;
			}

		if (!joueur_type[joueur])			// Si c'est  un humain de jouer...
		{
			do								// Dplacement du curseur jusqu' l'appui sur ENTER sur une case valide
			{
				draw_curseur(caze);
				touche = ngetchx();
				draw_curseur(caze);
				caze += (touche == HAUT)?((caze >= 9)?(-9):(63)):((touche == BAS)?((caze <= 62)?(9):(-63)):((touche == GAUCHE)?((caze%9 >= 1)?(-1):(7)):((touche == DROITE)?((caze%9 <= 6)?(1):(-7)):(0))));
				if ((touche == DELETE)&&(nbre_pions > 4))						// Pour revenir en arrire d'un coup
				{
					nbre_pions--;
					if (joueur_type[!joueur])
					{
						nbre_pions--;
						DrawChar(90, 36, '', 2); DrawChar(90, 50, '', 2);
					}
					charger(nbre_pions, tableau, &joueur, &nbre_blancs);
					for (i=0; i<=70; i++) if (i%9 != 8) draw_pion(i, ((tableau[i] <= 1)?(tableau[i]):(11)));	// Rafrachit le tableau
					goto Boucle;
				}
				if (touche == PLUS) for (++caze%71; !test_case(tableau, caze, joueur); caze = ++caze%71);	// Pour passer  la case jouable suivante
				if (touche == CLEAR) off();
				if (touche == ESCAPE) menu(tableau, joueur_type, &nbre_ordi, &joueur, &nbre_blancs, 1);		// Accs au menu
			} while ((touche != ENTER)||(!test_case(tableau, caze, joueur)));
		}

		else											// Si c'est la 89 qui joue...
		{
			if (kbhit() == ESCAPE) menu(tableau, joueur_type, &nbre_ordi, &joueur, &nbre_blancs, 1); GKeyFlush();
			if ((joueur_type[joueur] == 1)||(nbre_pions == 4)||(nbre == 1)) { caze = tourner[random(nbre)]; goto Retourner; }
			DrawLine(10*(caze%9)+5, 10*caze/9-10*caze%9+23,10*(caze%9)+13, 10*caze/9-10*caze%9+23, 1);
			barre2 = 94;
			FontSetSys(PETIT);											// Initialisation de la barre de progression
			DrawClipRect(&(WIN_RECT){94, 70, 145, 78} , &(SCR_RECT){0, 0, 159, 99}, 1);
			DrawStr(95, 72, fr_eng("En reflexion...", "Thinking..."), 1);
			FontSetSys(NORMAL);
			if (nbre_pions > 56-joueur_type[joueur])					// Rflexion spciale pour la fin de jeu
			{
				max = -64;
				for (touche=0; touche<nbre; touche++)
				{
					memcpy(tab, tableau, sizeof(TABLEAU));
					for (horiz=((caz = tourner[touche])%9>=2)?(-1):(0); horiz<=(caz%9<=5)?(1):(0); horiz++)	// Pour chaque case, retourne les pions
						for (vert=(caz>=18)?(-1):(0); vert<=(caz<=52)?(1):(0); vert++)						// et lance la rflexion rcursive
						{
							pas = 9*vert+horiz;
							if (tab[caz+pas] != !joueur) continue;
							for (i=caz+2*pas; (pas>0)?(i<=70):(i>=0); i+=pas)
							{
								if (tab[i] > 1) break;
								if (tab[i] == joueur)
								{
									for (j=caz+pas; (pas>0)?(j<=i-pas):(j>=i-pas); j+=pas) tab[j] = joueur;
									break;
								}
							}
						}
					if (kbhit() == ESCAPE) menu(tableau, joueur_type, &nbre_ordi, &joueur, &nbre_blancs, 1);
					if (kbhit() == CLEAR) off();
					GKeyFlush();
					barre1 = barre2+1; barre2 = 95+(touche+1)*49/nbre;
					for (i=71; i<=77; i++) DrawLine(barre1, i, barre2, i, 2);
					tab[caz] = joueur;
					for (i=0, score=0; i<=59; i++) if (tab[ordre_nouveau[i]] > 1) ordre[score++] = ordre_nouveau[i];
					for (i=0, score=0; i<=70; i++) score += (tab[i] == joueur)-(tab[i] == !joueur);
					score = meilleur_coup_fin(tab, !joueur, nbre_pions, -max, -score, 0, ((nbre_pions > 59-joueur_type[joueur])?(-64):(-1)));
					if (score > max) { max = score; caze = caz; if ((max >= 1)&&(nbre_pions <= 59-joueur_type[joueur])) goto Retourner; }
				}
			}
			else														// Rflexion standard du jeu
			{
				max = -32000; 
				for (touche=0; touche<nbre; touche++)
				{
					memcpy(tab, tableau, sizeof(TABLEAU));
					for (horiz=((caz = tourner[touche])%9>=2)?(-1):(0); horiz<=(caz%9<=5)?(1):(0); horiz++)
						for (vert=(caz>=18)?(-1):(0); vert<=(caz<=52)?(1):(0); vert++)
						{
							pas = 9*vert+horiz;
							if (tab[caz+pas] != !joueur) continue;
							for (i=caz+2*pas; (pas>0)?(i<=70):(i>=0); i+=pas)
							{
								if (tab[i] > 1) break;
								if (tab[i] == joueur)
								{
									for (j=caz+pas; (pas>0)?(j<=i-pas):(j>=i-pas); j+=pas) tab[j] = joueur;
									break;
								}
							}
						}
					if (kbhit() == ESCAPE) menu(tableau, joueur_type, &nbre_ordi, &joueur, &nbre_blancs, 1);
					if (kbhit() == CLEAR) off();
					GKeyFlush();
					barre1 = barre2+1; barre2 = 95+(touche+1)*49/nbre;
					for (i=71; i<=77; i++) DrawLine(barre1, i, barre2, i, 2);
					tab[caz] = joueur;
					for (i=0, score=0; i<=59; i++) if (tab[ordre_nouveau[i]] > 1) ordre[score++] = ordre_nouveau[i];
					for (i=0, score=0; i<=70; i++) score += ((tab[i] == joueur)-(tab[i] == !joueur))*coeff[i];
					if (joueur_type[joueur] > 2)
					{
						for (horiz=(caz%9>=1)?(-1):(0); horiz<=(caz%9<=6)?(1):(0); horiz++)
							for (vert=(caz>=9)?(-1):(0); vert<=(caz<=61)?(1):(0); vert++) if (tab[caz+9*vert+horiz] == 3) tab[caz+9*vert+horiz] = 2; 
						score = meilleur_coup(tab, !joueur, joueur_type[joueur]+((nbre_pions < 56-2*joueur_type[joueur])?(-3):(-1)), -max, - score)-450*(1-2*(joueur_type[joueur]%2))+random(35);
					}
					else score += coeff[caz];
					if (score > max) { max = score; caze = caz; }
				}
			}
 		}

		Retourner:
		LCD_restore(LCD_tmp);
		ScrRectFill(&(SCR_RECT){102, 89, 150, 94}, &(SCR_RECT){0, 0, 159, 99}, 0);

		if ((joueur_type[joueur] > 1)&&(nbre > 0)&&tips)		// Affichage de l'valuation de la position
		{
			FontSetSys(PETIT);
			printf_xy(102, 89, "eval: %d", max);
			FontSetSys(NORMAL);
		}

		tab_sav[nbre_pions-4] = caze; tab_sav[nbre_pions-3] = 255;
		draw_pion(caze, joueur);
		tableau[caze] = joueur; score = 0;
		if (!joueur) nbre_blancs++;
		for (horiz=(caze%9>=2)?(-1):(0); horiz<=(caze%9<=5)?(1):(0); horiz++)		// Retourne les pions sans les afficher
			for (vert=(caze>=18)?(-1):(0); vert<=(caze<=52)?(1):(0); vert++)
			{
				pas = 9*vert+horiz;
				if (!(tableau[caze+pas] == !joueur)) continue;
				for (i=caze+2*pas; (pas>0)?(i<=70):(i>=0); i+=pas)
				{
					if (tableau[i] >= 2) break;
					if (tableau[i] == joueur)
					{
						for (j=caze+pas; (pas>0)?(j<=i-pas):(j>=i-pas); j+=pas)
						{
							tableau[j] = joueur; tourner[score++] = j;
							nbre_blancs += (joueur)?(-1):(1);
							if (!faire_tourner) draw_pion(j, joueur);
						}
						break;
					}
				}
			}
		if (faire_tourner)
		{
			for ((joueur)?(i=5):(i=2); (joueur)?(i>=1):(i<=6); (joueur)?(i--):(i++))	// Affiche les diffrentes tapes de
			{																			// rotation des pions
				for (j=0; j<score; j++) draw_pion(tourner[j], i);
				for (touche=0; touche<28000; touche++);
			}
		}

		evol_tableau(tableau, caze);

	}

	tab_sav[0] = 0;			// Pour qu'on ne puisse pas charger cette partie qui est finie
	FontSetSys(GRAS);
	ClrScr();
	DrawStr(fr_eng(4, 7), 30, (2*nbre_blancs > nbre_pions)?(fr_eng("Les blancs gagnent!", "White player wins!")):((2*nbre_blancs < nbre_pions)?(fr_eng("Les noirs gagnent!", "Black player wins!")):(fr_eng("     Ex aequo!", "       Tie!"))), 2);
	printf_xy(fr_eng(43, 40), 48, fr_eng("%d noirs", "%d blacks"), nbre_pions-nbre_blancs);
	printf_xy(40, 61, fr_eng("%d blancs", "%d whites"), nbre_blancs);
	FontSetSys(NORMAL);
	GKeyFlush(); ngetchx();
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Procdure appele au lancement
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void _main(void)
{
	int i, j;
	ULONG apd = OSTimerCurVal(2);

	if(!setjmp(jmp_quit)) while (1)
	{
		for (i=0; i<=49; i++)			// Je n'ai pas pu m'empcher de placer qqs jolis effets graphiques...
		{
			DrawClipRect(MakeWinRect(49-i, 49-i, 110+i, 50+i) , &(SCR_RECT){0, 0, 159, 99}, 1);
			DrawClipRect(MakeWinRect(50-i, 50-i, 109+i, 49+i) , &(SCR_RECT){0, 0, 159, 99}, 0);
			for (j=0; j<=4000; j++);
		}
		jeu();
	}

	OSVFreeTimer(2);
	OSRegisterTimer (2, apd);			// Restaure le timer APD sauv dans apd
	for (i=49; i>=0; i--)
	{
		draw_rect(49-i, 49-i, 110+i, 50+i, 1);
		for (j=0; j<=1300; j++);
	}
	GrayMode(GRAY_OFF);
}

